/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/* 
 * ezfc-alias.c
 * Copyright (C) 2011 Akira TAGOH
 * 
 * Authors:
 *   Akira TAGOH  <akira@tagoh.org>
 * 
 * This library is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either
 * version 3 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <glib.h>
#include <libeasyfc/ezfc-error.h>
#include "ezfc-mem.h"
#include <libeasyfc/ezfc-alias.h>

struct _ezfc_alias_t {
	ezfc_mem_t  parent;
	gchar      *alias_name;
	FcPattern  *pattern;
};

/*< private >*/

/*< public >*/
ezfc_alias_t *
ezfc_alias_new(const gchar *alias_name)
{
	ezfc_alias_t *retval = ezfc_mem_alloc_object(sizeof (ezfc_alias_t));

	if (retval) {
		retval->alias_name = g_strdup(alias_name);
		ezfc_mem_add_ref(&retval->parent, retval->alias_name, g_free);
		retval->pattern = NULL;
	}

	return retval;
}

ezfc_alias_t *
ezfc_alias_ref(ezfc_alias_t *alias)
{
	g_return_val_if_fail (alias != NULL, NULL);

	return ezfc_mem_ref(&alias->parent);
}

void
ezfc_alias_unref(ezfc_alias_t *alias)
{
	if (alias)
		ezfc_mem_unref(&alias->parent);
}

const gchar *
ezfc_alias_get_name(ezfc_alias_t *alias)
{
	g_return_val_if_fail (alias != NULL, NULL);

	return alias->alias_name;
}

FcPattern *
ezfc_alias_get_font_pattern(ezfc_alias_t *alias)
{
	g_return_val_if_fail (alias != NULL, NULL);

	return FcPatternDuplicate(alias->pattern);
}

gboolean
ezfc_alias_set_font_pattern(ezfc_alias_t     *alias,
			    const FcPattern  *pattern,
			    GError          **error)
{
	FcPattern *pat;
	gboolean retval = TRUE;
	GError *err = NULL;

	g_return_val_if_fail (alias != NULL, FALSE);
	g_return_val_if_fail (pattern != NULL, FALSE);

	pat = FcPatternDuplicate(pattern);
	if (!pat) {
		g_set_error(&err, EZFC_ERROR, EZFC_ERR_OOM,
			    "Unable to make a duplicate of FcPattern");
		retval = FALSE;
		goto bail;
	}
	if (alias->pattern)
		ezfc_mem_remove_ref(&alias->parent, alias->pattern);
	alias->pattern = pat;
	ezfc_mem_add_ref(&alias->parent, pat, (ezfc_destroy_func_t)FcPatternDestroy);

  bail:
	if (err) {
		if (error)
			*error = g_error_copy(err);
		else
			g_warning(err->message);
		g_error_free(err);
	}

	return retval;
}

const gchar *
ezfc_alias_get_font(ezfc_alias_t *alias)
{
	gchar *family = NULL;

	g_return_val_if_fail (alias != NULL, NULL);

	if (alias->pattern &&
	    FcPatternGetString(alias->pattern, FC_FAMILY, 0, (FcChar8 **)&family) != FcResultMatch)
		return NULL;

	return family;
}

gboolean
ezfc_alias_set_font(ezfc_alias_t  *alias,
		    const gchar   *font_name,
		    GError       **error)
{
	FcPattern *pat, *match = NULL;
	FcResult result;
	gboolean retval = TRUE;
	GError *err = NULL;

	g_return_val_if_fail (alias != NULL, FALSE);
	g_return_val_if_fail (font_name != NULL, FALSE);

	pat = FcNameParse((FcChar8 *)font_name);
	if (!pat) {
		g_set_error(&err, EZFC_ERROR, EZFC_ERR_FAIL_ON_FC,
			    "Unable to parse `%s'",
			    font_name);
		retval = FALSE;
		goto bail;
	}
	if ((match = FcFontMatch(NULL, pat, &result))) {
		FcChar8 *v1, *v2;

		if (FcPatternGetString(pat, FC_FAMILY, 0, &v1) != FcResultMatch) {
			g_set_error(&err, EZFC_ERROR, EZFC_ERR_NO_FAMILY,
				    "Pattern doesn't contain any family name");
			retval = FALSE;
			goto bail1;
		}
		if (FcPatternGetString(match, FC_FAMILY, 0, &v2) != FcResultMatch) {
			g_set_error(&err, EZFC_ERROR, EZFC_ERR_NO_FAMILY,
				    "Pattern doesn't contain any family name");
			retval = FALSE;
			goto bail1;
		}
		if (FcStrCmpIgnoreCase(v1, v2) == 0) {
			retval = ezfc_alias_set_font_pattern(alias, pat, &err);
		} else {
			g_set_error(&err, EZFC_ERROR, EZFC_ERR_NO_VALID_FONT,
				    "No such fonts available: %s",
				    font_name);
			retval = FALSE;
			goto bail1;
		}
	  bail1:
		FcPatternDestroy(match);
	} else {
		g_set_error(&err, EZFC_ERROR, EZFC_ERR_NO_VALID_FONT,
			    "No such fonts available: %s",
			    font_name);
		retval = FALSE;
		goto bail;
	}
  bail:
	FcPatternDestroy(pat);

	if (err) {
		if (error)
			*error = g_error_copy(err);
		else
			g_warning(err->message);
		g_error_free(err);
	}

	return retval;
}
