/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/* 
 * ezfc-config.c
 * Copyright (C) 2011 Akira TAGOH
 * 
 * Authors:
 *   Akira TAGOH  <akira@tagoh.org>
 * 
 * This library is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either
 * version 3 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <glib.h>
#include <libxml/tree.h>
#include <libxml/xpath.h>
#include <libeasyfc/ezfc-error.h>
#include "ezfc-mem.h"
#include <libeasyfc/ezfc-config.h>

struct _ezfc_config_t {
	ezfc_mem_t  parent;
	GHashTable *aliases;
};
typedef enum {
	EZFC_NODE_TYPE_PREFER = 0,
	EZFC_NODE_TYPE_ACCEPT,
	EZFC_NODE_TYPE_DEFAULT,
	EZFC_NODE_TYPE_END
} ezfc_node_t;

/*< private >*/
static void
_ezfc_alias_list_free(gpointer p)
{
	GList *l = p, *ll;

	for (ll = l; ll != NULL; ll = g_list_next(ll)) {
		ezfc_alias_unref(ll->data);
	}
	g_list_free(l);
}

static xmlNodePtr
_ezfc_config_to_xml_node(const gchar *language,
			 const gchar *test_font,
			 const gpointer     edit_object,
			 ezfc_node_t  type)
{
	xmlNodePtr match, test, edit;
	FcPattern *pat;
	FcChar8 *family;
	const gchar *mode[] = {
		"prepend",
		"append",
		"append_last",
		NULL
	};

	g_return_val_if_fail (type < EZFC_NODE_TYPE_END, NULL);

	match = xmlNewNode(NULL,
			   (const xmlChar *)"match");
	xmlNewProp(match,
		   (const xmlChar *)"target",
		   (const xmlChar *)"pattern");

	test = xmlNewChild(match, NULL,
			   (const xmlChar *)"test",
			   NULL);
	xmlNewProp(test,
		   (const xmlChar *)"name",
		   (const xmlChar *)"lang");
	xmlNewChild(test, NULL,
		    (const xmlChar *)"string",
		    (const xmlChar *)language);

	test = xmlNewChild(match, NULL,
			   (const xmlChar *)"test",
			   NULL);
	xmlNewProp(test,
		   (const xmlChar *)"name",
		   (const xmlChar *)"family");
	xmlNewChild(test, NULL,
		    (const xmlChar *)"string",
		    (const xmlChar *)test_font);

	if (type == EZFC_NODE_TYPE_DEFAULT) {
		family = edit_object;
	} else {
		pat = edit_object;
		/* XXX: there are no way to peek everything in FcPattern efficiently.
		   so only support family name so far.
		*/
		if (FcPatternGetString(pat, FC_FAMILY, 0, &family) != FcResultMatch)
			return NULL;
	}
	edit = xmlNewChild(match, NULL,
			   (const xmlChar *)"edit",
			   NULL);
	xmlNewProp(edit,
		   (const xmlChar *)"name",
		   (const xmlChar *)"family");
	xmlNewProp(edit,
		   (const xmlChar *)"mode",
		   (const xmlChar *)mode[type]);
	xmlNewChild(edit, NULL,
		    (const xmlChar *)"string",
		    (const xmlChar *)family);

	return match;
}

static gboolean
_ezfc_config_to_xml(ezfc_config_t  *config,
		    xmlChar       **output,
		    int            *size)
{
	xmlDocPtr doc;
	xmlNodePtr root;
	gchar *comment;
	GHashTableIter iter;
	gboolean no_elements = TRUE;
	gpointer key, val;

	doc = xmlNewDoc((const xmlChar *)"1.0");
	doc->encoding = xmlStrdup((const xmlChar *)"UTF-8");
	xmlCreateIntSubset(doc,
			   (const xmlChar *)"fontconfig",
			   NULL,
			   (const xmlChar *)"../fonts.dtd");
	root = xmlNewDocNode(doc,
			     NULL,
			     (const xmlChar *)"fontconfig",
			     NULL);
	xmlDocSetRootElement(doc, root);
	comment = g_strdup_printf("\n"
				  "\tTHIS FILE WAS GENERATED BY libeasyfc %s\n"
				  "\tDO NOT EDIT THIS DIRECTLY\n"
				  "\tANY CHANGES HAs BEEN MADE MANUALLY MAY BE LOST.\n",
				  VERSION);
	xmlAddChild(root, xmlNewComment((const xmlChar *)comment));
	g_free(comment);

	g_hash_table_iter_init(&iter, config->aliases);
	while (g_hash_table_iter_next(&iter, &key, &val)) {
		const gchar *slang = key;
		const GList *list = val, *l;
		xmlNodePtr node;

		for (l = list; l != NULL; l = g_list_next(l)) {
			ezfc_alias_t *a = l->data;
			const gchar *aliases[] = {
				"sans",
				"sans-serif",
				"serif",
				"monospace",
				"cursive",
				"fantasy",
				NULL
			}, *n;
			int i;
			gboolean is_subst = TRUE;
			FcPattern *pat;
			FcChar8 *family;

			n = ezfc_alias_get_name(a);
			pat = ezfc_alias_get_font_pattern(a);
			for (i = 0; aliases[i] != NULL; i++) {
				if (g_ascii_strcasecmp(n, aliases[i]) == 0) {
					is_subst = FALSE;
					break;
				}
			}
			if (is_subst) {
				node = _ezfc_config_to_xml_node(slang,
								n, pat,
								EZFC_NODE_TYPE_ACCEPT);
				xmlAddChild(root, node);
			} else {
				if (FcPatternGetString(pat, FC_FAMILY, 0, &family) != FcResultMatch)
					return FALSE;
				node = _ezfc_config_to_xml_node(slang,
								(const gchar *)family,
								(const gpointer)n,
								EZFC_NODE_TYPE_DEFAULT);
				xmlAddChild(root, node);
				node = _ezfc_config_to_xml_node(slang,
								n, pat,
								EZFC_NODE_TYPE_PREFER);
				xmlAddChild(root, node);
			}
			no_elements = FALSE;
			FcPatternDestroy(pat);
		}
	}

	xmlDocDumpFormatMemory(doc, output, size, 1);

	xmlFreeDoc(doc);

	return !no_elements;
}

static xmlChar *
_ezfc_config_parse_string_node(xmlNodePtr node)
{
	xmlChar *retval = NULL;

	while (node != NULL) {
		if (node->type == XML_COMMENT_NODE || node->type == XML_TEXT_NODE)
			node = node->next;
		if (node->type == XML_ELEMENT_NODE &&
		    xmlStrcmp(node->name, (const xmlChar *)"string") == 0) {
			retval = xmlNodeGetContent(node);
			break;
		} else {
			g_warning("Unexpected node: type = %d, name = %s",
				  node->type, node->name);
			break;
		}
	}

	return retval;
}

static gboolean
_ezfc_config_from_xml(ezfc_config_t  *config,
		      const gchar    *conffile,
		      GError        **error)
{
	xmlDocPtr doc = NULL;
	xmlXPathContextPtr xctxt = NULL;
	xmlXPathObjectPtr xobj = NULL;
	gboolean retval = TRUE;
	GError *err = NULL;
	int i, n;

	xmlInitParser();

	doc = xmlParseFile(conffile);
	if (!doc) {
		g_set_error(&err, EZFC_ERROR, EZFC_ERR_FAIL_ON_XML,
			    "Unable to read the configuration.");
		retval = FALSE;
		goto bail;
	}
	xctxt = xmlXPathNewContext(doc);
	if (!xctxt) {
		g_set_error(&err, EZFC_ERROR, EZFC_ERR_FAIL_ON_XML,
			    "Unable to create a XPath context.");
		retval = FALSE;
		goto bail;
	}
	xobj = xmlXPathEvalExpression((const xmlChar *)"/fontconfig/match[(@target = \"pattern\" or not(@*)) and (test[@name = \"lang\"] and test/string/text()) and (test[@name = \"family\"] and test/string/text()) and (edit[@name = \"family\"] and edit/string/text())]", xctxt);
	if (!xobj) {
		g_set_error(&err, EZFC_ERROR, EZFC_ERR_FAIL_ON_XML,
			    "No valid elements can be proceeded in this library.");
		retval = FALSE;
		goto bail;
	}

	n = xmlXPathNodeSetGetLength(xobj->nodesetval);

	for (i = 0; i < n; i++) {
		xmlNodePtr match = xmlXPathNodeSetItem(xobj->nodesetval, i);
		xmlNodePtr node;
		xmlChar *attr, *lang = NULL, *alias = NULL, *s;
		FcPattern *pat = FcPatternCreate();
		ezfc_alias_t *a;

		node = match->children;
		while (node != NULL) {
			if (node->type == XML_TEXT_NODE) {
				/* ignore the text node here */
				goto bail1;
			}
			if (xmlStrcmp(node->name, (const xmlChar *)"test") == 0) {
				attr = xmlGetProp(node, (const xmlChar *)"name");
				if (xmlStrcmp(attr, (const xmlChar *)"lang") == 0) {
					if (lang)
						xmlFree(lang);
					lang = _ezfc_config_parse_string_node(node->children);
				} else if (xmlStrcmp(attr, (const xmlChar *)"family") == 0) {
					if (alias)
						xmlFree(alias);
					alias = _ezfc_config_parse_string_node(node->children);
				} else {
					g_warning("Unexpected value in the name attribute on <test>: %s\n",
						  attr);
				}
				xmlFree(attr);
			} else if (xmlStrcmp(node->name, (const xmlChar *)"edit") == 0) {
				attr = xmlGetProp(node, (const xmlChar *)"mode");
				if (xmlStrcmp(attr, (const xmlChar *)"append_last") == 0) {
					/* kind of the alternatives of <default> */
					/* We are assuming that there should be the <prefer> rule too.
					 * so ignoring.
					 */
					xmlFree(attr);
					goto bail2;
				} else {
					xmlFree(attr);
					attr = xmlGetProp(node, (const xmlChar *)"name");
					if (xmlStrcmp(attr, (const xmlChar *)"family") == 0) {
						s = _ezfc_config_parse_string_node(node->children);
						FcPatternAddString(pat, FC_FAMILY, (const FcChar8 *)s);
						xmlFree(s);
					} else {
						g_warning("Unexpected value in the name attribute: %s",
							  attr);
					}
				}
				xmlFree(attr);
			} else {
				g_warning("Unexpected element in <match>: %s\n", node->name);
			}
		  bail1:
			node = node->next;
		}
		if (!lang) {
			g_warning("No language testing in the rule.");
			goto bail2;
		}
		if (!alias) {
			g_warning("No alias name is defined.");
			goto bail2;
		}
		a = ezfc_alias_new((const gchar *)alias);
		if (!ezfc_alias_set_font_pattern(a, pat, &err)) {
			ezfc_alias_unref(a);
			goto bail2;
		}
		ezfc_config_add_alias(config, (const gchar *)lang, a);
		ezfc_alias_unref(a);
	  bail2:
		if (lang)
			xmlFree(lang);
		if (alias)
			xmlFree(alias);
		if (pat)
			FcPatternDestroy(pat);
		if (err)
			break;
	}

  bail:
	if (err) {
		if (error)
			*error = g_error_copy(err);
		else
			g_warning(err->message);
		g_error_free(err);
	}
	if (xobj)
		xmlXPathFreeObject(xobj);
	if (xctxt)
		xmlXPathFreeContext(xctxt);
	if (doc)
		xmlFreeDoc(doc);
	xmlCleanupParser();

	return retval;
}

/*< public >*/
ezfc_config_t *
ezfc_config_new(void)
{
	ezfc_config_t *retval = ezfc_mem_alloc_object(sizeof (ezfc_config_t));

	if (retval) {
		retval->aliases = g_hash_table_new_full(g_str_hash,
							g_str_equal,
							g_free,
							_ezfc_alias_list_free);
		ezfc_mem_add_ref(&retval->parent, retval->aliases,
				 (ezfc_destroy_func_t)g_hash_table_destroy);
	}

	return retval;
}

ezfc_config_t *
ezfc_config_ref(ezfc_config_t *config)
{
	g_return_val_if_fail (config != NULL, NULL);

	return ezfc_mem_ref(&config->parent);
}

void
ezfc_config_unref(ezfc_config_t *config)
{
	if (config)
		ezfc_mem_unref(&config->parent);
}

gboolean
ezfc_config_add_alias(ezfc_config_t *config,
		      const gchar   *language,
		      ezfc_alias_t  *alias)
{
	GList *list, *l;

	g_return_val_if_fail (config != NULL, FALSE);
	g_return_val_if_fail (language != NULL, FALSE);
	g_return_val_if_fail (alias != NULL, FALSE);

	list = g_hash_table_lookup(config->aliases, language);
	if (!list) {
		list = g_list_append(list, ezfc_alias_ref(alias));
		g_hash_table_insert(config->aliases, g_strdup(language), list);
	} else {
		l = list;
		while (1) {
			ezfc_alias_t *a = l->data;

			if (g_ascii_strcasecmp(ezfc_alias_get_name(a),
					       ezfc_alias_get_name(alias)) == 0) {
				ezfc_alias_unref(a);
				l->data = ezfc_alias_ref(alias);
				goto bail;
			}
			if (g_list_next(l) == NULL)
				break;
			l = g_list_next(l);
		}
		l = g_list_append(l, ezfc_alias_ref(alias));
	}
  bail:

	return TRUE;
}

gboolean
ezfc_config_remove_alias(ezfc_config_t *config,
			 const gchar   *language,
			 const gchar   *alias_name)
{
	GList *list, *l;

	g_return_val_if_fail (config != NULL, FALSE);
	g_return_val_if_fail (language != NULL, FALSE);
	g_return_val_if_fail (alias_name != NULL, FALSE);

	if ((list = g_hash_table_lookup(config->aliases, language)) == NULL)
		return FALSE;

	for (l = list; l != NULL; l = g_list_next(l)) {
		ezfc_alias_t *a = l->data;
		gboolean need_replace = FALSE;

		if (g_ascii_strcasecmp(ezfc_alias_get_name(a),
				       alias_name) == 0) {
			if (l == list)
				need_replace = TRUE;
			ezfc_alias_unref(a);
			l = g_list_delete_link(l, l);
			if (need_replace)
				g_hash_table_replace(config->aliases,
						     g_strdup(language),
						     l);

			return TRUE;
		}
	}

	return FALSE;
}

/**
 * ezfc_config_get_language_list:
 * @config: self
 *
 *
 * Obtains the list of languages in #ezfc_config_t instance.
 *
 * Returns: (element-type utf8) (transfer none) (allow-none): a #GList contains languages or %NULL.
 */
GList *
ezfc_config_get_language_list(ezfc_config_t *config)
{
	g_return_val_if_fail (config != NULL, NULL);

	return g_hash_table_get_keys(config->aliases);
}

/**
 * ezfc_config_get_aliases:
 * @config: self
 * @language: a language name referenced to the alias
 *
 * Obtains the list of #ezfc_alias_t in #ezfc_config_t instance.
 *
 * Returns: (element-type Easyfc.Alias) (transfer container) (allow-none): a #GList contains #ezfc_alias_t or %NULL.
 */
GList *
ezfc_config_get_aliases(ezfc_config_t *config,
			const gchar   *language)
{
	g_return_val_if_fail (config != NULL, NULL);

	return g_hash_table_lookup(config->aliases, language);
}

gboolean
ezfc_config_load(ezfc_config_t  *config,
		 GError        **error)
{
	const gchar *configdir = g_get_user_config_dir();
	gchar *fcconfdir = g_build_filename(configdir, "fonts", NULL);
	gchar *fcconf = g_build_filename(fcconfdir, "ezfc.conf", NULL);
	gboolean retval = TRUE;
	GError *err = NULL;

	g_return_val_if_fail (config != NULL, FALSE);

	if (!g_file_test(fcconf, G_FILE_TEST_IS_REGULAR)) {
		g_set_error(&err, EZFC_ERROR, EZFC_ERR_NO_CONFIG_FILE,
			    "No such config file available: %s",
			    fcconf);
		retval = FALSE;
		goto bail;
	}
	if (!(retval = _ezfc_config_from_xml(config, fcconf, &err)))
		goto bail;

  bail:
	if (err) {
		if (error)
			*error = g_error_copy(err);
		else
			g_warning(err->message);
		g_error_free(err);
	}
	g_free(fcconf);
	g_free(fcconfdir);

	return retval;
}

gboolean
ezfc_config_save(ezfc_config_t  *config,
		 GError        **error)
{
	xmlChar *xml;
	int size;
	const gchar *homedir = g_get_home_dir();
	gchar *fcconfdir = g_build_filename(homedir, ".fonts.conf.d", NULL);
	gchar *ezfcconf = g_build_filename(fcconfdir, "00-ezfc.conf", NULL);
	GError *err = NULL;
	gboolean retval = TRUE;

	g_return_val_if_fail (config != NULL, FALSE);

	if (!g_file_test(fcconfdir, G_FILE_TEST_EXISTS)) {
		g_mkdir_with_parents(fcconfdir, 0755);
	} else {
		if (!g_file_test(fcconfdir, G_FILE_TEST_IS_DIR)) {
			g_set_error(&err, EZFC_ERROR, EZFC_ERR_NO_CONFIG_DIR,
				    "Unable to create a config dir: %s",
				    fcconfdir);
			retval = FALSE;
			goto bail;
		}
	}
	if (!(retval = _ezfc_config_to_xml(config, &xml, &size))) {
		g_set_error(&err, EZFC_ERROR, EZFC_ERR_NO_ELEMENTS,
			    "No configuration to output");
		goto bail;
	}

	if (!g_file_set_contents(ezfcconf, (const gchar *)xml, size, &err))
		goto bail;
  bail:
	if (err) {
		if (error)
			*error = g_error_copy(err);
		else
			g_warning(err->message);
		g_error_free(err);
	}
	g_free(ezfcconf);
	g_free(fcconfdir);
	xmlFree(xml);

	return retval;
}

void
ezfc_config_dump(ezfc_config_t *config)
{
	GHashTableIter iter;
	gpointer key, val;

	g_return_if_fail (config != NULL);

	g_hash_table_iter_init(&iter, config->aliases);
	while (g_hash_table_iter_next(&iter, &key, &val)) {
		const gchar *lang = key;
		GList *l = val;

		g_print("%s:\n", lang);
		for (; l != NULL; l = g_list_next(l)) {
			ezfc_alias_t *a = l->data;
			FcChar8 *result;
			FcPattern *pat;

			pat = ezfc_alias_get_font_pattern(a);
			result = FcPatternFormat(pat,
						 (FcChar8 *)"%{=unparse}");
			if (!result)
				g_print("EEEEE\n");
			g_print("\t%s: %s\n",
				ezfc_alias_get_name(a),
				result);
			FcStrFree(result);
			FcPatternDestroy(pat);
		}
	}
}
